#!/usr/bin/env python
"""
Module containing the tools for data generation
"""
from __future__ import division, print_function

import sys
import urllib
import urllib.request

import matplotlib
import numpy as np
import gwpy

matplotlib.use("agg")  # noqa
import bilby

from bilby_pipe.utils import logger, BilbyPipeError
from bilby_pipe.main import DataDump, parse_args
from bilby_pipe.input import Input
from bilby_pipe.parser import create_parser

try:
    import nds2  # noqa
except ImportError:
    logger.warning(
        "You do not have nds2 (python-nds2-client) installed. You may "
        " experience problems accessing interferometer data."
    )

try:
    import LDAStools  # noqa
except ImportError:
    logger.warning(
        "You do not have LDAStools (python-ldas-tools-framecpp) installed."
        " You may experience problems accessing interferometer data."
    )


class DataGenerationInput(Input):
    """ Handles user-input and creation of intermediate interferometer list

    Parameters
    ----------
    parser: configargparse.ArgParser, optional
        The parser containing the command line / ini file inputs
    args_list: list, optional
        A list of the arguments to parse. Defauts to `sys.argv[1:]`
    create_data: bool
        If false, no data is generated (used for testing)

    """

    def __init__(self, args, unknown_args, create_data=True):

        np.random.seed(args.generation_seed)

        logger.info("Command line arguments: {}".format(args))
        logger.info("Unknown command line arguments: {}".format(unknown_args))
        self.meta_data = dict(
            command_line_args=args,
            unknown_command_line_args=unknown_args,
            injection_parameters=None,
        )
        self.ini = args.ini
        self.cluster = args.cluster
        self.process = args.process
        self.idx = args.idx
        self.detectors = args.detectors
        self.channel_type = args.channel_type
        self.duration = args.duration
        self.post_trigger_duration = args.post_trigger_duration
        self.sampling_frequency = args.sampling_frequency
        self.psd_duration = args.psd_duration
        self.psd_start_time = args.psd_start_time
        self.psd_files = args.psd_files
        self.minimum_frequency = args.minimum_frequency
        self.outdir = args.outdir
        self.label = args.label
        self.frequency_domain_source_model = args.frequency_domain_source_model
        self.waveform_approximant = args.waveform_approximant
        self.reference_frequency = args.reference_frequency
        if create_data:
            self.create_data(args)

    def create_data(self, args):
        """ Function to iterarate through data generation method

        Note, the data methods are mutually exclusive and only one can given to
        the parser.

        Parameters
        ----------
        args: Namespace
            Input arguments

        Raises
        ------
        BilbyPipeError:
            If no data is generated

        """

        self.data_set = False
        # The following are all mutually exclusive methods to set the data
        if args.injection_file is not None:
            self.injection_file = args.injection_file
            self._set_interferometers_from_injection()
        elif self.data_set is False and args.gracedb is not None:
            self.gracedb = args.gracedb
            self._set_interferometers_from_data()
        elif self.data_set is False and args.gps_file is not None:
            self.gps_file = args.gps_file
            self._set_interferometers_from_data()
        elif self.data_set is False and args.trigger_time is not None:
            self.trigger_time = args.trigger_time
            self._set_interferometers_from_data()

        if self.data_set is False:
            raise BilbyPipeError("No data setting method provided")

    @property
    def cluster(self):
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        try:
            self._cluster = int(cluster)
        except (ValueError, TypeError):
            logger.debug("Unable to convert input `cluster` to type int")
            self._cluster = cluster

    @property
    def process(self):
        return self._process

    @process.setter
    def process(self, process):
        try:
            self._process = int(process)
        except (ValueError, TypeError):
            logger.debug("Unable to convert input `process` to type int")
            self._process = process

    @property
    def psd_duration(self):
        return self._psd_duration

    @psd_duration.setter
    def psd_duration(self, psd_duration):
        if psd_duration is None:
            self._psd_duration = 32 * self.duration
        else:
            self._psd_duration = psd_duration
        logger.info("PSD duration set to {}".format(self.psd_duration))

    @property
    def psd_start_time(self):
        if self._psd_start_time is not None:
            return self._psd_start_time
        elif self.trigger_time is not None:
            psd_start_time = self.start_time - self.psd_duration
            logger.info("Using default PSD start time {}".format(psd_start_time))
            return psd_start_time
        else:
            raise BilbyPipeError("PSD start time not set")

    @psd_start_time.setter
    def psd_start_time(self, psd_start_time):
        if psd_start_time is None:
            self._psd_start_time = None
        else:
            self._psd_start_time = psd_start_time
            logger.info("PSD start-time set to {}".format(self._psd_start_time))

    @property
    def minimum_frequency(self):
        return self._minimum_frequency

    @minimum_frequency.setter
    def minimum_frequency(self, minimum_frequency):
        self._minimum_frequency = float(minimum_frequency)

    @property
    def parameter_conversion(self):
        if "binary_neutron_star" in self.frequency_domain_source_model:
            return bilby.gw.conversion.convert_to_lal_binary_neutron_star_parameters
        elif "binary_black_hole" in self.frequency_domain_source_model:
            return bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters
        else:
            return None

    @property
    def detectors(self):
        """ A list of the detectors to search over, e.g., ['H1', 'L1'] """
        return self._detectors

    @detectors.setter
    def detectors(self, detectors):
        """ Handles various types of user input """
        if isinstance(detectors, list):
            if len(detectors) == 1:
                det_list = self._convert_string_to_list(detectors[0])
            else:
                det_list = detectors
        else:
            raise ValueError("Input `detectors` = {} not understood".format(detectors))

        det_list.sort()
        det_list = [det.upper() for det in det_list]
        self._detectors = det_list

    @property
    def trigger_time(self):
        return self._trigger_time

    @trigger_time.setter
    def trigger_time(self, trigger_time):
        self._trigger_time = trigger_time

    @property
    def gracedb(self):
        """ The gracedb of the candidate """
        return self._gracedb

    @gracedb.setter
    def gracedb(self, gracedb):
        """ Set the gracedb ID

        At setting, will load the json candidate data and path to the frame
        cache file.

        Parameters
        ----------
        gracedb: str
            The gracedb UID string

        """
        if gracedb is None:
            self._gracedb = None
        else:
            logger.info("Setting gracedb id to {}".format(gracedb))
            self.test_connection()
            candidate = bilby.gw.utils.gracedb_to_json(
                gracedb, outdir=self.data_directory
            )
            self.meta_data["gracedb_candidate"] = candidate
            self._gracedb = gracedb
            self.trigger_time = candidate["gpstime"]

    def test_connection(self):
        """ A generic test to see if the network is reachable """
        try:
            urllib.request.urlopen("https://google.com", timeout=0.1)
        except urllib.error.URLError:
            raise BilbyPipeError(
                "It appears you are not connected to a network and so won't be "
                "able to interface with GraceDB. You may wish to specify the "
                " local-generation argument either in the configuration file "
                "or by passing the --local-generation command line argument"
            )

    def _parse_gps_file(self):
        """ Reads in the GPS file selects the required time and set the trigger time

        Note, the gps_file setter method is defined in bilby_pipe.input
        """
        gps_start_times = self.read_gps_file()
        self.start_time = gps_start_times[self.idx]
        self.trigger_time = self.start_time + self.duration / 2.0

    def _set_psds_from_files(self, ifos):
        psd_file_dict = {}
        for psd_file in self.psd_files:
            ifo_name, file_path = psd_file.split(":")
            psd_file_dict[ifo_name] = file_path
        for ifo in [ifo for ifo in ifos if ifo.name in psd_file_dict.keys()]:
            ifo.power_spectral_density = bilby.gw.detector.PowerSpectralDensity.from_power_spectral_density_file(
                psd_file=psd_file_dict[ifo.name]
            )

    def _set_interferometers_from_injection(self):
        """ Method to generate the interferometers data from an injection """

        self.injection_parameters = self.injection_df.iloc[self.idx].to_dict()
        self.meta_data["injection_parameters"] = self.injection_parameters
        self.trigger_time = self.injection_parameters["geocent_time"]

        waveform_arguments = dict(
            waveform_approximant=self.waveform_approximant,
            reference_frequency=self.reference_frequency,
            minimum_frequency=self.minimum_frequency,
        )

        waveform_generator = bilby.gw.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=self.bilby_frequency_domain_source_model,
            parameter_conversion=self.parameter_conversion,
            waveform_arguments=waveform_arguments,
        )

        ifos = bilby.gw.detector.InterferometerList(self.detectors)

        if self.psd_files is not None:
            self._set_psds_from_files(ifos)

        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency,
            duration=self.duration,
            start_time=self.trigger_time - self.duration / 2,
        )

        ifos.inject_signal(
            waveform_generator=waveform_generator, parameters=self.injection_parameters
        )

        self.interferometers = ifos

    def _set_interferometers_from_data(self):
        """ Method to generate the interferometers data from data"""
        end_time = self.start_time + self.duration
        ifo_list = []
        for det in self.detectors:
            logger.info("Getting analysis-segment data for {}".format(det))
            data = self._get_data(det, self.channel_type, self.start_time, end_time)
            ifo = bilby.gw.detector.get_empty_interferometer(det)
            ifo.strain_data.set_from_gwpy_timeseries(data)

            psd_end_time = self.psd_start_time + self.psd_duration
            logger.info("Getting psd-segment data for {}".format(det))
            psd_data = self._get_data(
                det, self.channel_type, self.psd_start_time, psd_end_time
            )
            roll_off = 0.2
            psd_alpha = 2 * roll_off / self.duration
            logger.info(
                "Tukey window PSD data with alpha={}, roll off={}".format(
                    psd_alpha, roll_off
                )
            )
            psd = psd_data.psd(
                fftlength=self.duration, overlap=0, window=("tukey", psd_alpha)
            )
            ifo.power_spectral_density = bilby.gw.detector.PowerSpectralDensity(
                frequency_array=psd.frequencies.value, psd_array=psd.value
            )
            ifo_list.append(ifo)
        self.interferometers = bilby.gw.detector.InterferometerList(ifo_list)

    def _get_data(self, det, channel_type, start_time, end_time):
        """ Read in data using gwpy

        This first uses the `gwpy.timeseries.TimeSeries.get()` method to acces
        the data, if this fails, it then attempts to use `fetch_open_data()` as
        a fallback.

        Parameters
        ----------
        channel_type: str, or list of strings
            The full channel name is formed from <det>:<channel_type>, see
            bilby_pipe_generation --help for more information. If given as a
            list each type will be tried and the first success returned.
        start_time, end_time: float
            GPS start and end time of segment
        """
        data = None
        for ct in list(channel_type):
            channel = "{}:{}".format(det, ct)
            try:
                data = self._gwpy_get(det, channel, start_time, end_time)
                break
            except RuntimeError as e:
                logger.info("Unable to read data for channel {}".format(channel))
                logger.debug("Error message {}".format(e))
            except ImportError:
                logger.info("Unable to read data as NDS2 is not installed")
        if data is None:
            logger.warning(
                "Attempts to download data failed, trying with `fetch_open_data`"
            )
            data = gwpy.timeseries.TimeSeries.fetch_open_data(det, start_time, end_time)

        return data

    def _gwpy_get(self, det, channel, start_time, end_time):
        logger.info(
            "Calling TimeSeries.get({}, start_time={}, end_time={})".format(
                channel, start_time, end_time
            )
        )
        data = gwpy.timeseries.TimeSeries.get(
            channel, start_time, end_time, verbose=True
        )
        return data

    @property
    def interferometers(self):
        """ A bilby.gw.detector.InterferometerList """
        try:
            return self._interferometers
        except AttributeError:
            raise ValueError(
                "interferometers unset, did you provide a set-data method?"
            )

    @interferometers.setter
    def interferometers(self, interferometers):
        self._interferometers = interferometers
        self.data_set = True

    def save_interferometer_list(self):
        """ Method to dump the saved data to disk for later analysis """
        data_dump = DataDump(
            outdir=self.data_directory,
            label=self.label,
            idx=self.idx,
            trigger_time=self.trigger_time,
            interferometers=self.interferometers,
            meta_data=self.meta_data,
        )
        data_dump.to_pickle()


def create_generation_parser():
    return create_parser(
        pipe_args=False,
        job_args=True,
        run_spec=True,
        pe_summary=False,
        injection=True,
        data_gen=True,
        waveform=True,
        generation=True,
        analysis=False,
    )


def main():
    args, unknown_args = parse_args(sys.argv[1:], create_generation_parser())
    data = DataGenerationInput(args, unknown_args)
    data.save_interferometer_list()
    if args.create_plots:
        data.interferometers.plot_data(outdir=data.data_directory, label=data.label)



from bilby_pipe.bilbyargparser import BilbyArgParser
from . import utils

__version__ = utils.get_version_information()


def create_parser(
    pipe_args=True,
    job_args=True,
    run_spec=True,
    pe_summary=True,
    injection=True,
    data_gen=True,
    waveform=True,
    generation=True,
    analysis=True,
):
    """ Creates the BilbyArgParser for bilby_pipe

    Parameters
    ----------
    run_spec: bool
        Add the `run_spec` argument group to the parser
    pe_summary: bool
        Add the `pe_summary` argument group to the parser
    injection: bool
        Add the `injection` argument group to the parser
    data_gen: bool
        Add the `data_gen` argument group to the parser
    waveform: bool
        Add the `waveform` argument group to the parser
    detector: bool
        Add the `detector` argument group to the parser

    Returns
    -------
    parser: BilbyArgParser instance
        Argument parser

    """
    parser = BilbyArgParser(
        usage=__doc__, ignore_unknown_config_file_keys=True, allow_abbrev=False
    )
    parser.add("ini", type=str, is_config_file=True, help="Configuration ini file")
    parser.add("-v", "--verbose", action="store_true", help="Verbose output")
    parser.add(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )

    if job_args:
        parser.add("--idx", type=int, help="The level A job index", default=0)
        parser.add("--cluster", type=str, help="The condor cluster ID", default=None)
        parser.add("--process", type=str, help="The condor process ID", default=None)
        parser.add(
            "--create-plots",
            action="store_true",
            help="Create diagnostic and posterior plots",
        )

    if pipe_args:
        parser.add(
            "--accounting", type=str, required=True, help="Accounting group to use"
        )
        parser.add(
            "--local",
            action="store_true",
            help="Run the job locally, i.e., not through a batch submission",
        )
        parser.add(
            "--local-generation",
            action="store_true",
            help=(
                "Run the data generation job locally. Note that if you are "
                "running on a cluster where the compute nodes do not have "
                "internet access, e.g. on ARCCA, you will need to run the "
                "data generation job locally."
            ),
        )
        parser.add(
            "--submit",
            action="store_true",
            help="Attempt to submit the job after the build",
        )
        parser.add(
            "--X509",
            type=str,
            default=None,
            help=(
                "The path to the users X509 certificate file."
                " Defaults to a copy of the file at the environment variable"
                " $X509_USER_PROXY saved in outdir and linked in"
                " the condor jobs submission"
            ),
        )
        parser.add(
            "--singularity-image",
            type=str,
            default=None,
            help="Singularity image to use",
        )
        parser.add(
            "--request-memory",
            type=float,
            default=4,
            help="Memory allocation request (GB)",
        )
        parser.add("--request-cpus", type=int, default=1, help="CPU allocation request")

    if run_spec:
        parser.add("--label", type=str, default="label", help="Output label")
        parser.add("--outdir", type=str, default=".", help="Output directory")
        parser.add(
            "--sampler",
            nargs="+",
            default="dynesty",
            help="Sampler to use, or a list of samplers to use",
        )
        parser.add(
            "--detectors",
            action="append",
            help=(
                "The names of detectors to use. If given in the ini file, "
                "detectors are specified by `detectors=[H1, L1]`. If given "
                "at the command line, as `--detectors H1 --detectors L1`"
            ),
        )
        parser.add(
            "--coherence-test",
            action="store_true",
            help=(
                "Run the analysis for all detectors together and for each "
                "detector separately"
            ),
        )

    if pe_summary:
        pe_summary = parser.add_argument_group(title="Summary page arguments")
        pe_summary.add(
            "--create-summary", action="store_true", help="Create a summary page"
        )
        pe_summary.add(
            "--webdir",
            type=str,
            default=None,
            help=(
                "Directory to store summary pages. If not given, defaults to "
                "outdir/results_page"
            ),
        )
        pe_summary.add("--email", type=str, help="Email for notifications")
        pe_summary.add(
            "--existing-dir",
            type=str,
            default=None,
            help=(
                "If given, add results to an directory with an an existing"
                " summary.html file"
            ),
        )

    if injection:
        injection_parser = parser.add_argument_group(title="Injection arguments")
        injection_parser.add(
            "--injection",
            action="store_true",
            default=False,
            help="Create data from an injection file",
        )
        injection_parser.add(
            "--injection-file",
            type=str,
            default=None,
            help="Injection file: overrides `n-injection`.",
        )
        injection_parser.add_arg(
            "--n-injection",
            type=int,
            help="Number of injections to generate by sampling from the prior",
        )

    if data_gen:
        data_gen_pars = parser.add_mutually_exclusive_group()
        data_gen_pars.add(
            "--trigger-time",
            default=None,
            type=float,
            help="The trigger time, alternative to --gracedb",
        )
        data_gen_pars.add(
            "--gps-file",
            type=str,
            help="File containing segment GPS start times",
            default=None,
        )
        data_gen_pars.add("--gracedb", type=str, help="Gracedb UID", default=None)
        parser.add(
            "--psd-files",
            default=None,
            nargs="*",
            help="PSD files to use, if not provided known PSDs will be used",
        )

    if waveform:
        parser.add(
            "--reference-frequency",
            default=20,
            type=float,
            help="The reference frequency",
        )
        parser.add(
            "--waveform-approximant",
            default="IMRPhenomPv2",
            type=str,
            help="The name of the waveform approximant",
        )
        parser.add(
            "--frequency-domain-source-model",
            default="lal_binary_black_hole",
            type=str,
            help=(
                "Name of the frequency domain source model. Can be one of"
                "[lal_binary_black_hole, lal_binary_neutron_star,"
                "lal_eccentric_binary_black_hole_no_spins, sinegaussian, "
                "supernova, supernova_pca_model]"
            ),
        )

    if generation:
        det_parser = parser.add_argument_group(title="Detector arguments")
        det_parser.add(
            "--duration",
            type=int,
            default=4,
            help="The duration of data around the event to use",
        )
        det_parser.add(
            "--post-trigger-duration",
            type=int,
            default=2,
            help=(
                "Used to set the segment start in relation to the trigger time."
                " The segment start time is given by "
                " trigger_time + post_trigger_duration - duration"
            ),
        )
        det_parser.add("--sampling-frequency", default=2048, type=int)
        det_parser.add(
            "--channel-type",
            nargs="*",
            default=["DCH-CLEAN_STRAIN_C02"],
            help=(
                "Channel type to use. Typical options are [DCH-CLEAN_STRAIN_C02, "
                "DCS-CALIB_STRAIN_C02, DCS-CALIB_STRAIN_C01, GDS-CALIB_STRAIN]"
            ),
        )
        det_parser.add(
            "--psd-duration",
            default=None,
            type=float,
            help="Time used to generate the PSD, default is 32x the sample duration.",
        )
        det_parser.add(
            "--psd-start-time",
            default=None,
            type=float,
            help=(
                "Start time of data used to generate the PSD. If not given, "
                "defaults to psd-duration/2 before the trigger time"
            ),
        )
        det_parser.add("--minimum-frequency", default=20, type=float)
        det_parser.add(
            "--generation-seed",
            default=None,
            type=int,
            help="Random seed used during data generation",
        )

    if analysis:
        parser.add(
            "--data-label",
            default=None,
            help="Label used for the data dump",
            required=True,
        )
        parser.add(
            "--sampling-seed", default=None, type=int, help="Random sampling seed"
        )
        parser.add(
            "--default-prior",
            default="BBHPriorDict",
            type=str,
            help="The name of the prior set to base the prior on. Can be one of"
            "[PriorDict, BBHPriorDict, BNSPriorDict, CalibrationPriorDict]",
        )
        parser.add("--prior-file", default=None, help="The prior file")
        parser.add(
            "--likelihood-type",
            default="GravitationalWaveTransient",
            help="The likelihood. Can be one of"
            "[GravitationalWaveTransient, ROQGravitationalWaveTransient]"
            "Need to specify --roq-folder if ROQ likelihood used",
        )
        parser.add("--roq-folder", default=None, help="The data for ROQ")
        parser.add(
            "--deltaT",
            type=float,
            default=0.1,
            help=(
                "The symmetric width (in s) around the trigger time to"
                " search over the coalesence time"
            ),
        )
        parser.add(
            "--distance-marginalization",
            action="store_true",
            default=False,
            help="Boolean. If true, use a distance-marginalized likelihood",
        )
        parser.add(
            "--phase-marginalization",
            action="store_true",
            default=False,
            help="Boolean. If true, use a phase-marginalized likelihood",
        )
        parser.add(
            "--time-marginalization",
            action="store_true",
            default=False,
            help="Boolean. If true, use a time-marginalized likelihood",
        )
        parser.add("--sampler-kwargs", default=None)

    return parser



#!/usr/bin/env python
"""
Script to analyse the stored data
"""
from __future__ import division, print_function

import sys
import os

import numpy as np
import matplotlib

matplotlib.use("agg")  # noqa
import bilby

from bilby_pipe.utils import logger, BilbyPipeError
from bilby_pipe.main import DataDump, parse_args
from bilby_pipe.parser import create_parser
from bilby_pipe.input import Input


class DataAnalysisInput(Input):
    """ Handles user-input and analysis of intermediate ifo list

    Parameters
    ----------
    parser: BilbyArgParser, optional
        The parser containing the command line / ini file inputs
    args_list: list, optional
        A list of the arguments to parse. Defauts to `sys.argv[1:]`

    """

    def __init__(self, args, unknown_args):
        logger.info("Command line arguments: {}".format(args))

        self.ini = args.ini
        self.idx = args.idx
        self.cluster = args.cluster
        self.process = args.process
        self.detectors = args.detectors
        self.prior_file = args.prior_file
        self._priors = None
        self.deltaT = args.deltaT
        self.reference_frequency = args.reference_frequency
        self.waveform_approximant = args.waveform_approximant
        self.distance_marginalization = args.distance_marginalization
        self.phase_marginalization = args.phase_marginalization
        self.time_marginalization = args.time_marginalization
        self.sampling_seed = args.sampling_seed
        self.sampler = args.sampler
        self.sampler_kwargs = args.sampler_kwargs
        self.outdir = args.outdir
        self.label = args.label
        self.data_label = args.data_label
        self.default_prior = args.default_prior
        self.frequency_domain_source_model = args.frequency_domain_source_model
        self.likelihood_type = args.likelihood_type
        self.roq_folder = args.roq_folder
        self.result = None

    @property
    def cluster(self):
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        try:
            self._cluster = int(cluster)
        except (ValueError, TypeError):
            logger.debug("Unable to convert input `cluster` to type int")
            self._cluster = cluster

    @property
    def process(self):
        return self._process

    @process.setter
    def process(self, process):
        try:
            self._process = int(process)
        except (ValueError, TypeError):
            logger.debug("Unable to convert input `process` to type int")
            self._process = process

    @property
    def reference_frequency(self):
        return self._reference_frequency

    @reference_frequency.setter
    def reference_frequency(self, reference_frequency):
        self._reference_frequency = float(reference_frequency)

    @property
    def sampling_seed(self):
        return self._samplng_seed

    @sampling_seed.setter
    def sampling_seed(self, sampling_seed):
        if sampling_seed is None:
            sampling_seed = np.random.randint(1, 1e6)
        self._samplng_seed = sampling_seed
        np.random.seed(sampling_seed)
        logger.info("Sampling seed set to {}".format(sampling_seed))

    @property
    def sampler(self):
        return self._sampler

    @sampler.setter
    def sampler(self, sampler):
        """ Setter for the sampler

        By default, the input parser takes a list of samplers (to enable DAGs
        to be specified to run over all the samplers). In the analysis script,
        it must be given only a single sampler. This logic checks that is the
        case and raises an error otherwise
        """
        if isinstance(sampler, str):
            self._sampler = sampler
        elif isinstance(sampler, list):
            if len(sampler) == 1:
                self._sampler = sampler[0]
            else:
                raise BilbyPipeError(
                    "Data analsys script recieved a list of samplers with "
                    "more than one element: {}. Unable to proceed".format(sampler)
                )

    @property
    def sampler_kwargs(self):
        return self._sampler_kwargs

    @sampler_kwargs.setter
    def sampler_kwargs(self, sampler_kwargs):
        if sampler_kwargs is not None:
            try:
                self._sampler_kwargs = eval(sampler_kwargs)
            except (NameError, TypeError) as e:
                raise BilbyPipeError(
                    "Error {}. Unable to parse sampler_kwargs: {}".format(
                        e, sampler_kwargs
                    )
                )
        else:
            self._sampler_kwargs = None

    @property
    def interferometers(self):
        try:
            return self._interferometers
        except AttributeError:
            ifos = self.data_dump.interferometers
            names = [ifo.name for ifo in ifos]
            logger.info("Found data for detectors = {}".format(names))
            ifos_to_use = [ifo for ifo in ifos if ifo.name in self.detectors]
            names_to_use = [ifo.name for ifo in ifos_to_use]
            logger.info("Using data for detectors = {}".format(names_to_use))
            self._interferometers = bilby.gw.detector.InterferometerList(ifos_to_use)
            self.print_detector_information(self._interferometers)
            return self._interferometers

    @staticmethod
    def print_detector_information(interferometers):
        for ifo in interferometers:
            logger.info(
                "{}: sampling-frequency={}, segment-start-time={}, duration={}".format(
                    ifo.name,
                    ifo.strain_data.sampling_frequency,
                    ifo.strain_data.start_time,
                    ifo.strain_data.duration,
                )
            )

    @property
    def meta_data(self):
        return self.data_dump.meta_data

    @property
    def trigger_time(self):
        return self.data_dump.trigger_time

    @property
    def data_dump(self):
        try:
            return self._data_dump
        except AttributeError:
            filename = DataDump.get_filename(
                self.data_directory, self.data_label, str(self.idx)
            )
            self._data_dump = DataDump.from_pickle(filename)
            return self._data_dump

    @property
    def priors(self):
        if self._priors is None:
            if self.default_prior in bilby.core.prior.__dict__.keys():
                self._priors = bilby.core.prior.__dict__[self.default_prior](
                    filename=self.prior_file
                )
            elif self.default_prior in bilby.gw.prior.__dict__.keys():
                self._priors = bilby.gw.prior.__dict__[self.default_prior](
                    filename=self.prior_file
                )
            else:
                logger.info("No prior {} found.").format(self.default_prior)
                logger.info("Defaulting to BBHPriorDict")
                self._priors = bilby.gw.prior.BBHPriorDict(filename=self.prior_file)
            if isinstance(
                self._priors, (bilby.gw.prior.BBHPriorDict, bilby.gw.prior.BNSPriorDict)
            ):
                self._priors["geocent_time"] = bilby.core.prior.Uniform(
                    minimum=self.trigger_time - self.deltaT / 2,
                    maximum=self.trigger_time + self.deltaT / 2,
                    name="geocent_time",
                    latex_label="$t_c$",
                    unit="$s$",
                )
        return self._priors

    @property
    def parameter_conversion(self):
        if "binary_neutron_star" in self._frequency_domain_source_model:
            return bilby.gw.conversion.convert_to_lal_binary_neutron_star_parameters
        elif "binary_black_hole" in self._frequency_domain_source_model:
            return bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters
        else:
            return None

    @property
    def waveform_generator(self):
        if self.likelihood_type == "GravitationalWaveTransient":
            waveform_generator = bilby.gw.WaveformGenerator(
                sampling_frequency=self.interferometers.sampling_frequency,
                duration=self.interferometers.duration,
                frequency_domain_source_model=self.bilby_frequency_domain_source_model,
                parameter_conversion=self.parameter_conversion,
                start_time=self.interferometers.start_time,
                waveform_arguments=self.waveform_arguments,
            )

        elif self.likelihood_type == "ROQGravitationalWaveTransient":
            logger.info(
                "Using the ROQ likelihood with roq-folder={}".format(self.roq_folder)
            )
            freq_nodes_linear = np.load(self.roq_folder + "/fnodes_linear.npy")
            freq_nodes_quadratic = np.load(self.roq_folder + "/fnodes_quadratic.npy")

            waveform_arguments = self.waveform_arguments.copy()
            waveform_arguments["frequency_nodes_linear"] = freq_nodes_linear
            waveform_arguments["frequency_nodes_quadratic"] = freq_nodes_quadratic

            waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
                sampling_frequency=self.interferometers.sampling_frequency,
                duration=self.interferometers.duration,
                frequency_domain_source_model=bilby.gw.source.roq,
                start_time=self.interferometers.start_time,
                parameter_conversion=self.parameter_conversion,
                waveform_arguments=waveform_arguments,
            )

        else:
            raise ValueError("Unknown likelihood function")

        return waveform_generator

    @property
    def waveform_arguments(self):
        return dict(
            reference_frequency=self.reference_frequency,
            waveform_approximant=self.waveform_approximant,
            minimum_frequency=self.interferometers[0].minimum_frequency,
        )  # FIXME

    @property
    def likelihood(self):
        if self.likelihood_type == "GravitationalWaveTransient":
            return bilby.gw.likelihood.GravitationalWaveTransient(
                interferometers=self.interferometers,
                waveform_generator=self.waveform_generator,
                priors=self.priors,
                phase_marginalization=self.phase_marginalization,
                distance_marginalization=self.distance_marginalization,
                time_marginalization=self.time_marginalization,
            )

        elif self.likelihood_type == "ROQGravitationalWaveTransient":
            logger.info(
                "Using the ROQ likelihood with roq-folder={}".format(self.roq_folder)
            )
            basis_matrix_linear = np.load(self.roq_folder + "/B_linear.npy").T
            basic_matrix_quadratic = np.load(self.roq_folder + "/B_quadratic.npy").T

            if self.time_marginalization:
                logger.warning(
                    "Time marginalization not implemented for "
                    "ROQGravitationalWaveTransient: option ignored"
                )
            return bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.interferometers,
                waveform_generator=self.waveform_generator,
                priors=self.priors,
                linear_matrix=basis_matrix_linear,
                quadratic_matrix=basic_matrix_quadratic,
                phase_marginalization=self.phase_marginalization,
                distance_marginalization=self.distance_marginalization,
            )

        else:
            raise ValueError("Unknown likelihood function")

    @property
    def parameter_generation(self):
        if self.likelihood_type == "ROQGravitationalWaveTransient":
            # FIXME this is temporary given that the SNR cannot be computed
            # for the roq source model, as it passes mode=linear to
            # antenna_detector_response
            return None
        else:
            if "binary_neutron_star" in self._frequency_domain_source_model:
                return bilby.gw.conversion.generate_all_bns_parameters
            elif "binary_black_hole" in self._frequency_domain_source_model:
                return bilby.gw.conversion.generate_all_bbh_parameters
            else:
                return None

    @property
    def result_class(self):
        """ The bilby result class to store results in """
        try:
            return bilby.gw.result.CompactBinaryCoalesenceResult
        except AttributeError:
            logger.debug("Unable to use CBC specific result class")
            return None

    def run_sampler(self):
        self.result = bilby.run_sampler(
            likelihood=self.likelihood,
            priors=self.priors,
            sampler=self.sampler,
            label=self.label,
            outdir=os.path.relpath(self.result_directory),
            conversion_function=self.parameter_generation,
            injection_parameters=self.data_dump.meta_data["injection_parameters"],
            result_class=self.result_class,
            **self.sampler_kwargs
        )


def create_analysis_parser():
    return create_parser(
        pipe_args=False,
        job_args=True,
        run_spec=True,
        pe_summary=False,
        injection=False,
        data_gen=False,
        waveform=True,
        generation=False,
        analysis=True,
    )


def main():
    args, unknown_args = parse_args(sys.argv[1:], create_analysis_parser())
    analysis = DataAnalysisInput(args, unknown_args)
    analysis.run_sampler()
    if args.create_plots:
        analysis.result.plot_corner()
        analysis.result.plot_marginals()



#!/usr/bin/env python
"""
Module containing the tools for creating injection files
"""
from __future__ import division, print_function

import sys
import json

import pandas as pd
import numpy as np
import matplotlib

matplotlib.use("agg")  # noqa
import bilby
from bilby.gw.prior import BBHPriorDict

from .input import Input
from .utils import parse_args, logger, BilbyPipeError
from .bilbyargparser import BilbyArgParser


def create_parser():
    """ Generate a parser for the create_injections.py script

    Additional options can be added to the returned parser beforing calling
    `parser.parse_args` to generate the arguments`

    Returns
    -------
    parser: BilbyArgParser
        A parser with all the default options already added

    """
    parser = BilbyArgParser(ignore_unknown_config_file_keys=True)
    parser.add("ini", type=str, is_config_file=True, help="The ini file")
    parser.add("--label", type=str, default="LABEL", help="The output label")
    parser.add(
        "--outdir", type=str, default="bilby_outdir", help="The output directory"
    )
    parser.add_arg(
        "--prior-file",
        type=str,
        default=None,
        help="The prior file from which to generate injections",
    )
    parser.add_arg(
        "--n-injection", type=int, help="The number of injections to generate"
    )
    parser.add(
        "--generation-seed",
        default=None,
        type=int,
        help="Random seed used during data generation",
    )

    return parser


class CreateInjectionInput(Input):
    """ An object to hold all the inputs to create_injection

    Parameters
    ----------
    parser: configargparse.ArgParser, optional
        The parser containing the command line / ini file inputs
    args_list: list, optional
        A list of the arguments to parse. Defauts to `sys.argv[1:]`

    """

    def __init__(self, args, unknown_args):
        np.random.seed(args.generation_seed)
        logger.debug("Creating new CreateInjectionInput object")
        logger.info("Command line arguments: {}".format(args))

        self.prior_file = args.prior_file
        self.n_injection = args.n_injection
        self.outdir = args.outdir
        self.label = args.label

    @property
    def n_injection(self):
        """ The number of injections to create """
        if self._n_injection is not None:
            return self._n_injection
        else:
            raise BilbyPipeError("The number of injection has not been set")

    @n_injection.setter
    def n_injection(self, n_injection):
        self._n_injection = n_injection

    @property
    def prior_file(self):
        return self._prior_file

    @prior_file.setter
    def prior_file(self, prior_file):
        if isinstance(prior_file, str):
            logger.debug("Setting prior_file to {}".format(prior_file))
            self._prior_file = prior_file
            logger.debug("Generating PriorDict")
            self.prior = BBHPriorDict(filename=self._prior_file)
        else:
            raise BilbyPipeError("Type of prior_file must be str")

    @property
    def prior(self):
        return self._prior

    @prior.setter
    def prior(self, prior):
        if isinstance(prior, bilby.core.prior.PriorDict):
            self._prior = prior
        else:
            raise BilbyPipeError("Input prior not understood")

    def create_injection_file(self, filename):
        logger.info(
            "Generating injection file with prior={}, n_injection={}".format(
                self.prior, self.n_injection
            )
        )
        injection_values = pd.DataFrame.from_dict(self.prior.sample(self.n_injection))
        injections = dict(injections=injection_values)
        with open(filename, "w") as file:
            json.dump(
                injections, file, indent=2, cls=bilby.core.result.BilbyResultJsonEncoder
            )
        logger.info("Created injection file {}".format(filename))


def main():
    args, unknown_args = parse_args(sys.argv[1:], create_parser())
    inputs = CreateInjectionInput(args, unknown_args)
    inputs.create_injection_file()
